
//		SMBStuff.cpp - return SMBIOS information


#define UNICODE						1
#define _UNICODE					1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR					0
#define VER_MINOR					26
#define VER_BUILD					3
#define VER_PATCH					1

#define PLUGIN_NAME					L"SMBStuff"

#ifdef _WIN64
#define PLATFORM					L"x64"
#else
#define PLATFORM					L"x86"
#endif


#define MAX_ARG_LEN					8192
#define MAX_ARGS					1024
#define SHORT_BUF_LEN				1024

#define MAX_STRUCTURES				1024

#define SMB_TYPE_BIOSINFO			0
#define SMB_TYPE_SYSINFO			1
#define SMB_TYPE_BASEBOARD			2
#define SMB_TYPE_CHASSIS			3
#define SMB_TYPE_CPU				4
#define SMB_TYPE_END_OF_TABLE		127

#define SMB_INFO_MANUFACTURER		0x04
#define SMB_INFO_PRODUCT_NAME		0x05
#define SMB_INFO_VERSION_NUMBER		0x06
#define SMB_INFO_SERIAL_NUMBER		0x07
#define SMB_INFO_ASSET_TAG			0x08
#define SMB_INFO_FAMILY				0x1a
#define SMB_INFO_CHASSIS_TYPE		0x05
#define SMB_INFO_BIOS_VENDOR		0x04
#define SMB_INFO_BIOS_VERSION		0x05
#define SMB_INFO_BIOS_DATE			0x08
#define SMB_INFO_BIOS_REV_MAJOR		0x14
#define SMB_INFO_BIOS_REV_MINOR		0x15
#define SMB_INFO_CPU_SOCKET			0x04
#define SMB_INFO_CPU_VERSION		0x10
#define SMB_INFO_CPU_MANUFACTURER	0x07
#define SMB_INFO_CPU_EXT_CLOCK		0x12
#define SMB_INFO_CPU_SPEED			0x16




LPTSTR arg[MAX_ARGS];
int arglen[MAX_ARGS];
int numargs = 0;
wchar_t allargs[MAX_ARG_LEN] = L"";


const wchar_t PluginFeatures[] = L"_smbassettag,_smbbbassettag,_smbbbmfg,_smbbbproduct,_smbbbserial,_smbbbversion,\
_smbbiosdate,_smbbiosrev,_smbbiosvendor,_smbbiosversion,\
_smbbox,_smbchassis,_smbchassislock,_smbchassistag,_smbchassistype,_smbcpuclock,_smbcpumfg,_smbcpusocket,_smbcpuspeed,\
_smbcpuversion,_smbfamily,_smbmfg,_smbproduct,_smbserial,_smbver,@randx,@smbbyte,@smbdword,@smbqword,@smbstring,@smbword";


const wchar_t HelpWindowTitle[] = L"SMBStuff Plugin Help";


/*
const wchar_t RemarkPurposeText[] = L"Display a trivial message.\n\n";

const wchar_t RemarkUsageText[] = L"REMARK text...\r\n\n\
Any actual text arguments will in fact be ignored.  \n";
*/

const wchar_t _SMBAssetTagHelpText[] = L"_SMBASSETTAG \r\n\n\
Returns the chassis asset tag string, or  \n\
failing that, the baseboard asset tag;\n\
or N/A if neither one is found.\n";

const wchar_t _SMBBBAssetTagHelpText[] = L"_SMBBBASSETTAG \r\n\n\
Returns the baseboard asset tag string, or \n\
N/A if it is not present.\n";

const wchar_t _SMBBBMfgHelpText[] = L"_SMBBBMFG \r\n\n\
Returns the baseboard manufacturer string, or \n\
N/A if it is not present.\n";

const wchar_t _SMBBBProductHelpText[] = L"_SMBBBPRODUCT \r\n\n\
Returns the baseboard product name string, or \n\
N/A if it is not present.\n";

const wchar_t _SMBBBSerialHelpText[] = L"_SMBBBSERIAL \r\n\n\
Returns the baseboard serial number string, or  \n\
N/A if it is not present.\n";

const wchar_t _SMBBBVersionHelpText[] = L"_SMBBBVERSION \r\n\n\
Returns the baseboard version string, or \n\
N/A if it is not present.\n";

const wchar_t _SMBBIOSDateHelpText[] = L"_SMBBIOSDATE \r\n\n\
Returns the system BIOS date string, or  \n\
N/A if it is not present.\n";

const wchar_t _SMBBIOSRevHelpText[] = L"_SMBBIOSREV \r\n\n\
Returns the system BIOS revision, X.y, or  \n\
N/A if it is not present.\n";

const wchar_t _SMBBIOSVendorHelpText[] = L"_SMBBIOSVENDOR \r\n\n\
Returns the system BIOS vendor string, or  \n\
N/A if it is not present.\n";

const wchar_t _SMBBIOSVersionHelpText[] = L"_SMBBIOSVERSION \r\n\n\
Returns the system BIOS version string, or  \n\
N/A if it is not present.\n";

const wchar_t _SMBBoxHelpText[] = L"_SMBBOX \r\n\n\
Returns a one-letter code for the chassis type:  \n\
  L    for Portable, LapTop, Notebook, Hand Held, or Sub Notebook  \n\
  D    for anything else.\n";

const wchar_t _SMBChassisHelpText[] = L"_SMBCHASSIS \r\n\n\
Returns the system chassis byte, or  \n\
N/A if it is not present.\n";

const wchar_t _SMBChassisLockHelpText[] = L"_SMBCHASSISLOCK \r\n\n\
Returns 'Present' or 'Not present', or  \n\
N/A if this information is not present.  \n";

const wchar_t _SMBChassisTagHelpText[] = L"_SMBCHASSISTAG \r\n\n\
Returns the chassis asset tag string, or  \n\
N/A if this information is not present.\n";

const wchar_t _SMBChassisTypeHelpText[] = L"_SMBCHASSISTYPE \r\n\n\
Returns a string describing the system chassis type, or  \n\
N/A if this information is not present.\n";

const wchar_t _SMBCPUClockHelpText[] = L"_SMBCPUCLOCK \r\n\n\
Returns the external processor clock speed in MHz, or  \n\
N/A if this information is not present.\n";

const wchar_t _SMBCPUMfgHelpText[] = L"_SMBCPUMFG \r\n\n\
Returns the processor manufacturer string, or  \n\
N/A if it is not present.\n";

const wchar_t _SMBCPUSocketHelpText[] = L"_SMBCPUSOCKET \r\n\n\
Returns a string describing the processor socket, or \n\
N/A if it is not present.\n";

const wchar_t _SMBCPUSpeedHelpText[] = L"_SMBCPUSPEED \r\n\n\
Returns a the processor speed in MHz at boot time, or  \n\
N/A if this information is not present.\n";

const wchar_t _SMBCPUVersionHelpText[] = L"_SMBCPUVERSION \r\n\n\
Returns the processor version string, or  \n\
N/A if it is not present.\n";

const wchar_t _SMBFamilyHelpText[] = L"_SMBFAMILY \r\n\n\
Returns the system family name string, or  \n\
N/A if it is not present.\n";

const wchar_t _SMBMfgHelpText[] = L"_SMBMFG \r\n\n\
Returns the system manufacturer string, or  \n\
N/A if it is not present.\n";

const wchar_t _SMBProductHelpText[] = L"_SMBPRODUCT \r\n\n\
Returns the system product name string, or  \n\
N/A if it is not present.\n";

const wchar_t _SMBSerialHelpText[] = L"_SMBSERIAL \r\n\n\
Returns the system serial number string, or  \n\
N/A if it is not present.\n";

const wchar_t _SMBVerHelpText[] = L"_SMBVER \r\n\n\
Returns the SMBIOS version, X.y, or  \n\
N/A if it is not available.\n";

const wchar_t F_RandXHelpText[] = L"@RANDX[n] \r\n\n\
Generates a string of n random hex digits.  \n";

const wchar_t F_SMBByteHelpText[] = L"@SMBBYTE[type,offset,handle] \r\n\n\
Returns a byte value from a SMB structure.  \n\
\n\
  type\t0 - 255\n\
  offset\t4 - 255\n\
  handle\t0 - 0xffff (optional)  \n\
\n\
Returns N/A on any error.\n";

const wchar_t F_SMBDWordHelpText[] = L"@SMBDWORD[type,offset,handle] \r\n\n\
Returns a double-word value from a SMB structure.  \n\
\n\
  type\t0 - 255\n\
  offset\t4 - 252\n\
  handle\t0 - 0xffff (optional)  \n\
\n\
Returns N/A on any error.\n";


const wchar_t F_SMBQWordHelpText[] = L"@SMBQWORD[type,offset,handle] \r\n\n\
Returns a quad-word value from a SMB structure.  \n\
\n\
  type\t0 - 255\n\
  offset\t4 - 248\n\
  handle\t0 - 0xffff (optional)  \n\
\n\
Returns N/A on any error.\n";


const wchar_t F_SMBStringHelpText[] = L"@SMBSTRING[type,offset,handle] \r\n\n\
Returns a string from a SMB structure.  \n\
\n\
  type\t0 - 255\n\
  offset\t4 - 255\n\
  handle\t0 - 0xffff (optional)  \n\
\n\
Returns N/A on any error.\n";

const wchar_t F_SMBWordHelpText[] = L"@SMBWORD[type,offset,handle] \r\n\n\
Returns a word value from a SMB structure.  \n\
\n\
  type\t0 - 255\n\
  offset\t4 - 254\n\
  handle\t0 - 0xffff (optional)  \n\
\n\
Returns N/A on any error.\n";


const int NumHelpIndexItems = 31;

struct _HelpIndexItem {
	wchar_t Name[16];
	LPCTSTR HelpText;
	LPCTSTR UsageText;
};

_HelpIndexItem HelpIndex[NumHelpIndexItems] = {
	{ L"_smbassettag",		_SMBAssetTagHelpText,		NULL },
	{ L"_smbbbassettag",	_SMBBBAssetTagHelpText,		NULL },
	{ L"_smbbbmfg",			_SMBBBMfgHelpText,			NULL },
	{ L"_smbbbproduct",		_SMBBBProductHelpText,		NULL },
	{ L"_smbbbserial",		_SMBBBSerialHelpText,		NULL },
	{ L"_smbbbversion",		_SMBBBVersionHelpText,		NULL },
	{ L"_smbbiosdate",		_SMBBIOSDateHelpText,		NULL },
	{ L"_smbbiosrev",		_SMBBIOSRevHelpText,		NULL },
	{ L"_smbbiosvendor",	_SMBBIOSVendorHelpText,		NULL },
	{ L"_smbbiosversion",	_SMBBIOSVersionHelpText,	NULL },
	{ L"_smbbox",			_SMBBoxHelpText,			NULL },
	{ L"_smbchassis",		_SMBChassisHelpText,		NULL },
	{ L"_smbchassislock",	_SMBChassisLockHelpText,	NULL },
	{ L"_smbchassistag",	_SMBChassisTagHelpText,		NULL },
	{ L"_smbchassistype",	_SMBChassisTypeHelpText,	NULL },
	{ L"_smbcpuclock",		_SMBCPUClockHelpText,		NULL },
	{ L"_smbcpumfg",		_SMBCPUMfgHelpText,			NULL },
	{ L"_smbcpusocket",		_SMBCPUSocketHelpText,		NULL },
	{ L"_smbcpuspeed",		_SMBCPUSpeedHelpText,		NULL },
	{ L"_smbcpuversion",	_SMBCPUVersionHelpText,		NULL },
	{ L"_smbfamily",		_SMBFamilyHelpText,			NULL },
	{ L"_smbmfg",			_SMBMfgHelpText,			NULL },
	{ L"_smbproduct",		_SMBProductHelpText,		NULL },
	{ L"_smbserial",		_SMBSerialHelpText,			NULL },
	{ L"_smbver",			_SMBVerHelpText,			NULL },
	{ L"@randx",			F_RandXHelpText,			NULL },
	{ L"@smbbyte",			F_SMBByteHelpText,			NULL },
	{ L"@smbdword",			F_SMBDWordHelpText,			NULL },
	{ L"@smbqword",			F_SMBQWordHelpText,			NULL },
	{ L"@smbstring",		F_SMBStringHelpText,		NULL },
	{ L"@smbword",			F_SMBWordHelpText,			NULL }
};



const wchar_t StrNotAvailable[] = L"N/A";



UINT FirmwareTableSize	= 0;
BYTE *FirmwareTable		= NULL;
BYTE SMBVersionLow = 0, SMBVersionHigh = 0;


struct SMBStructure {
	BYTE Type;
	BYTE Length;
	WORD Handle;
};

SMBStructure *StructurePtr[MAX_STRUCTURES];
unsigned int NumStructures = 0;



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( GetStdHandle(STD_ERROR_HANDLE), shortbuf );

	return 0;
}


BOOL ValidStructure( SMBStructure *Structure, SMBStructure **NextStructure )
{
	BYTE *b = NULL;
	//	int StringCount = 0;


	//		Check for null pointers:

	if ( !Structure || !FirmwareTable )
		return FALSE;
	
	//		Check for structure address too low:

	if ( Structure < (SMBStructure*) FirmwareTable )
		return FALSE;

	//		Check for valid length byte:

	if ( Structure->Length < 4 )
		return FALSE;

	//		Check for potential buffer overrun:

	if ( ( (BYTE*) Structure + Structure->Length + 1 ) > ( FirmwareTable + FirmwareTableSize ) )
		return FALSE;

	//		Search for the end of the structure:

	b = (BYTE*) Structure;
	b += Structure->Length;

	/*
	if ( *b )
		StringCount++;
	*/

	while ( *b | *(b + 1) ) {
		/*
		if ( *b == 0 )
			StringCount++;
		*/

		b++;
		if ( ( b + 1 ) >= ( FirmwareTable + FirmwareTableSize ) )
			return FALSE;
	}


	//	Printf( L"Found structure type %u, handle %04x, length %u, with %d strings, at address 0x%08x.\r\n", Structure->Type, Structure->Handle, Structure->Length, StringCount, Structure );

	if ( NextStructure ) {
		if ( Structure->Type == SMB_TYPE_END_OF_TABLE )
			*NextStructure = NULL;
		else if ( ( b + 2 ) >= ( FirmwareTable + FirmwareTableSize ) )
			*NextStructure = NULL;
		else
			*NextStructure = (SMBStructure*) ( b + 2 );
	}

	return TRUE;
}


int LoadFirmwareTable()
{
	typedef UINT ( WINAPI *GETSYSFIRMWARETABLE ) ( DWORD ProviderSignature, DWORD TableID, PVOID TableBuffer, DWORD BufferSize );
	GETSYSFIRMWARETABLE GetSysFirmwareTable = NULL;

	HMODULE Kernel32DllHandle = GetModuleHandle( L"Kernel32.dll" );

	GetSysFirmwareTable = ( GETSYSFIRMWARETABLE ) GetProcAddress( Kernel32DllHandle, "GetSystemFirmwareTable" );

	if ( !GetSysFirmwareTable ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"GetSystemFirmwareTable() not supported!\r\n" );
		return 1;
	}


	FirmwareTableSize = GetSysFirmwareTable( 'RSMB', 0x0000, NULL, 0x0000 );

	if ( !FirmwareTableSize ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"No firmware table!\r\n" );
		return 2;
	}


	FirmwareTable = (BYTE*) malloc( FirmwareTableSize );

	if ( !FirmwareTable ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Error allocating buffer for firmware table!\r\n" );
		return 3;
	}


	if ( !GetSysFirmwareTable( 'RSMB', 0x0000, FirmwareTable, FirmwareTableSize ) ) {
		free( FirmwareTable );
		FirmwareTable = NULL;
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Error loading firmware table!\r\n" );
		return 4;
	}


	SMBStructure *FirstStructure = (SMBStructure*) FirmwareTable;
	SMBStructure *ThisStructure = NULL;
	SMBStructure *NextStructure = NULL;
	NumStructures = 0;


	if ( FirmwareTableSize >= 14 ) {				//		8-byte Microsoft header:  Is this documented anywhere?
		if ( ( FirmwareTable[4] + 256 * FirmwareTable[5] + 8 ) == FirmwareTableSize ) {
			if ( ( FirmwareTable[1] >= 2 ) && ( FirmwareTable[1] <= 10 ) ) {
				SMBVersionHigh = FirmwareTable[1];
				SMBVersionLow  = FirmwareTable[2];
				FirstStructure = (SMBStructure*) ( FirmwareTable + 8 );
			}
		}
	}


	ThisStructure = FirstStructure;
	while ( ValidStructure( ThisStructure, &NextStructure ) && ( NumStructures < MAX_STRUCTURES ) ) {
		StructurePtr[NumStructures++] = ThisStructure;
		ThisStructure = NextStructure;
	}

	if ( !NumStructures ) {
		free( FirmwareTable );
		FirmwareTable = NULL;
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Firmware tables don't appear valid!\r\n" );
		return 5;
	}

	return 0;
}


SMBStructure *FindStructure( BYTE Type, int Handle )
{
	if ( !NumStructures )
		return NULL;

	if ( Handle > 0xffff )
		Handle = -1;

	for ( unsigned int i = 0; i < NumStructures; i++ )
		if ( ( StructurePtr[i]->Type == Type ) && ( ( Handle < 0 ) || ( StructurePtr[i]->Handle == Handle ) ) )
			return StructurePtr[i];

	return NULL;
}


int GetSMBString( SMBStructure *Structure, unsigned int Offset, LPTSTR String, unsigned int BufferSize )
{
	if ( !Structure || !String )
		return -1;

	if ( !ValidStructure( Structure, NULL ) )
		return 1;

	if ( ( Offset < 4 ) || ( Offset >= Structure->Length ) )
		return 2;


	wcscpy_s( String, BufferSize, StrNotAvailable );


	BYTE *index		= (BYTE*) Structure;
	index += Offset;

	if ( *index == 0 )
		return 0;

	BYTE *b			= (BYTE*) Structure;
	b += Structure->Length;

	BYTE i			= 1;
	unsigned int j	= 0;


	while ( i < *index ) {
		if ( *b == 0x00 )
			return 3;

		while ( *b )
			b++;

		b++;
		i++;
	}

	if ( strlen( (char*) b ) >= BufferSize )
		return 4;

	while ( *b ) {
		String[j++] = *b;
		b++;
	}

	String[j] = 0x00;
	return 0;
}



int SeedRandomGenerator( )
{
	DWORD q = 0, a = 0, b = 0;
	SYSTEMTIME time;

	GetSystemTime( &time );
	a = ( ( time.wHour << 16 ) | time.wMinute ) * 1997;
	b = ( ( time.wSecond << 16 ) | time.wMilliseconds ) * 449;
	q = ( ( time.wMonth << 16 ) | time.wDay ) * 17;
	q = q ^ a ^ b;

	srand( q );
	return 0;
}


BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}


DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/smbstuff.html";
	piInfo.pszDescription	= (LPTSTR) L"SMBStuff plugin";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}


DLLExports BOOL WINAPI InitializePlugin( void )
{
	if ( LoadFirmwareTable() ) {
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Plugin not loaded.\r\n" );
		return -1;
	}


	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%d.%d.%d loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	SeedRandomGenerator();
	return 0;
}


DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	if ( FirmwareTable )
		free( FirmwareTable );

	return 0;
}



//	My funky idiosyncratic argument parser.  Supports a bunch of oddball parsing options,
//	some of which are used in this plugin.

#define PARSE_BREAK_SPACES		1
#define PARSE_BREAK_COMMAS		2
#define PARSE_SLASHES_KLUDGE	4
#define	PARSE_QUOTES_KLUDGE		8
#define PARSE_EQUALS_KLUDGE		16
#define PARSE_ONE_ARG_KLUDGE	32
#define PARSE_RETAIN_QUOTES		64
#define PARSE_FORCE_UPPERCASE	128
#define PARSE_NO_TRIM			256
#define PARSE_IGNORE_QUOTES		512
#define PARSE_EQ_QT_KLUDGE		1024


int ParseArgs( LPTSTR inString, unsigned int flags )
{
	/*
		flags:  bit 0 - break args at spaces
				bit 1 - break args at commas (either bit 0 or bit 1 must be set)
				bit 2 - slashes kludge:   treat /a/b like /a /b
				bit 3 - quotes kludge:    treat /a"foo" like /a:"foo"
				bit 4 - equals kludge:    break at first unquoted equals sign
				bit 5 - one-arg kludge:   if arg does not begin with /, allow unquoted spaces
				bit 6 - don't swallow double quotes
				bit 7 - force uppercase
				bit 8 - don't trim spaces from the right
				bit 9 - disable all special handling of double quotes
	*/

	if ( !inString )
		return -1;

	numargs = 0;

	int argnum = 0;
	int i = 0;
	int a = 0;
	int b = 0;
	wchar_t ch = '\0';
	BOOL quote  = FALSE;
	BOOL adv    = FALSE;
	BOOL done   = FALSE;
	int eqkl    = 0;

	if ( ( flags & ( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_EQUALS_KLUDGE ) ) == 0 )
		flags |= PARSE_BREAK_SPACES;

	for ( i = 0; i < MAX_ARGS; i++ ) {
		arg[i] = NULL;
		arglen[i] = 0;
	}

	for ( i = 0; !done; i++ ) {
		ch = inString[i];
		adv = FALSE;

		if ( ( ch == '"' ) && ( ( flags & PARSE_IGNORE_QUOTES ) == 0 ) ) {
			quote = !quote;

			if ( quote && ( a == 2 ) && ( arg[argnum][0] == '/' ) && ( iswalpha( arg[argnum][1] ) ) && ( flags & PARSE_QUOTES_KLUDGE ) ) {
				if ( flags & PARSE_RETAIN_QUOTES ) {
					if ( b < ( MAX_ARG_LEN - 1 ) )
						allargs[b++] = ':';
				}
				else
					ch = ':';
			}
			else if ( ( flags & PARSE_RETAIN_QUOTES ) == 0 )
				continue;
		}

		if ( iswspace( ch ) && ( a == 0 ) && !quote )
			continue;

		if ( ( a == 0 ) && ( ch != '/' ) && ( ( flags & ( PARSE_BREAK_SPACES | PARSE_ONE_ARG_KLUDGE ) ) == ( PARSE_BREAK_SPACES | PARSE_ONE_ARG_KLUDGE ) ) )
			flags &= ~PARSE_BREAK_SPACES;

		if ( iswspace( ch ) && !quote && ( a == 1 ) && ( argnum > 0 ) && ( flags & PARSE_EQUALS_KLUDGE ) && ( arg[argnum][0] == '=' ) )
			continue;

		if ( iswspace( ch ) && !quote && ( flags & PARSE_BREAK_SPACES ) )
			adv = TRUE;

		if ( ( ch == ',' ) && !quote && ( flags & PARSE_BREAK_COMMAS ) )
			adv = TRUE;

		if ( ( ch == '/' ) && !quote && ( a > 1 ) && ( arg[argnum][0] == '/' ) && ( flags & PARSE_SLASHES_KLUDGE ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
		}

		if ( ( ch == '=' ) && !quote && ( eqkl == 0 ) && ( a > 0 ) && ( arg[argnum][0] != '/' ) && ( flags & PARSE_EQUALS_KLUDGE ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			eqkl = argnum;
			flags &= ~( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS );
			//- Printf( L"   * equals kludge invoked; eqkl = %d\n", eqkl );

			if ( flags & PARSE_EQ_QT_KLUDGE )
				flags |= PARSE_IGNORE_QUOTES;
		}
		else if ( ( ch == '=' ) && !quote && ( flags & PARSE_EQUALS_KLUDGE ) ) {
			eqkl = -1;
			flags &= ~( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS );

			if ( flags & PARSE_EQ_QT_KLUDGE )
				flags |= PARSE_IGNORE_QUOTES;
		}

		if ( ch == '\0' ) {
			if ( a )
				argnum++;

			done = TRUE;
			continue;
		}

		if ( adv ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			quote = FALSE;
			continue;
		}

		if ( b >= ( MAX_ARG_LEN - 1 ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Buffer overflow: b = %d, MAX_ARG_LEN = %d \r\n", b, MAX_ARG_LEN );
			return -1;
		}

		if ( argnum < MAX_ARGS ) {
			if ( a == 0 )
				arg[argnum] = allargs + b;

			if ( flags & PARSE_FORCE_UPPERCASE )
				ch = towupper( ch );

			a++;
			arglen[argnum]  = a;
			allargs[b++] = ch;
			allargs[b]   = '\0';
		}

	}

	if ( ( flags & PARSE_NO_TRIM ) == 0 ) {
		for ( i = 0; i < argnum; i++ ) {
			while ( ( arglen[i] > 0 ) && iswspace( arg[i][arglen[i] - 1] ) ) {
				arglen[i]--;
				arg[i][arglen[i]] = '\0';
			}
		}
	}

	numargs = argnum;
	return argnum;
}


int ParseInt( LPTSTR inString, unsigned int *outValue )
{
	if ( !inString || !outValue )
		return -1;

	unsigned long int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( *sep1 != '\0' )
			return -1;

		*outValue = value;
		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( *sep1 != '\0' )
			return -1;

		*outValue = value;
		return 0;
	}

	return -1;
}



DLLExports int WINAPI _smbmfg( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *SysInfo = FindStructure( SMB_TYPE_SYSINFO, -1 );

	GetSMBString( SysInfo, SMB_INFO_MANUFACTURER, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbproduct( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *SysInfo = FindStructure( SMB_TYPE_SYSINFO, -1 );

	GetSMBString( SysInfo, SMB_INFO_PRODUCT_NAME, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbserial( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *SysInfo = FindStructure( SMB_TYPE_SYSINFO, -1 );

	GetSMBString( SysInfo, SMB_INFO_SERIAL_NUMBER, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbfamily( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *SysInfo = FindStructure( SMB_TYPE_SYSINFO, -1 );

	GetSMBString( SysInfo, SMB_INFO_FAMILY, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbbbassettag( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *BaseboardInfo = FindStructure( SMB_TYPE_BASEBOARD, -1 );

	GetSMBString( BaseboardInfo, SMB_INFO_ASSET_TAG, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbbbmfg( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *BaseboardInfo = FindStructure( SMB_TYPE_BASEBOARD, -1 );

	GetSMBString( BaseboardInfo, SMB_INFO_MANUFACTURER, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbbbproduct( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *BaseboardInfo = FindStructure( SMB_TYPE_BASEBOARD, -1 );

	GetSMBString( BaseboardInfo, SMB_INFO_PRODUCT_NAME, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbbbserial( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *BaseboardInfo = FindStructure( SMB_TYPE_BASEBOARD, -1 );

	GetSMBString( BaseboardInfo, SMB_INFO_SERIAL_NUMBER, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbbbversion( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *BaseboardInfo = FindStructure( SMB_TYPE_BASEBOARD, -1 );

	GetSMBString( BaseboardInfo, SMB_INFO_VERSION_NUMBER, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbassettag( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *ChassisInfo	= FindStructure( SMB_TYPE_CHASSIS, -1 );
	SMBStructure *BaseboardInfo	= FindStructure( SMB_TYPE_BASEBOARD, -1 );

	GetSMBString( ChassisInfo, SMB_INFO_ASSET_TAG, outString, MAX_ARG_LEN );

	if ( ( outString[0] == '\0' ) || !wcscmp( outString, StrNotAvailable ) )
		GetSMBString( BaseboardInfo, SMB_INFO_ASSET_TAG, outString, MAX_ARG_LEN );

	return 0;
}


DLLExports int WINAPI _smbchassistag( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *ChassisInfo = FindStructure( SMB_TYPE_CHASSIS, -1 );

	GetSMBString( ChassisInfo, SMB_INFO_ASSET_TAG, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbbox( LPTSTR outString )
{
	if ( !outString )
		return -1;

	wcscpy_s( outString, MAX_ARG_LEN, L"D" );

	SMBStructure *SysInfo = FindStructure( SMB_TYPE_CHASSIS, -1 );
	if ( !SysInfo )
		return 0;

	if ( SysInfo->Length <= SMB_INFO_CHASSIS_TYPE )
		return 0;

	BYTE *b = (BYTE*) SysInfo;
	b += SMB_INFO_CHASSIS_TYPE;

	if ( ( *b == 0x08 ) || ( *b == 0x09 ) || ( *b == 0x0a ) || ( *b == 0x0b ) || ( *b == 0x0e ) )
		wcscpy_s( outString, MAX_ARG_LEN, L"L" );

	return 0;
}


DLLExports int WINAPI _smbchassis( LPTSTR outString )
{
	if ( !outString )
		return -1;

	wcscpy_s( outString, MAX_ARG_LEN, StrNotAvailable );

	SMBStructure *SysInfo = FindStructure( SMB_TYPE_CHASSIS, -1 );
	if ( !SysInfo )
		return 0;

	if ( SysInfo->Length <= SMB_INFO_CHASSIS_TYPE )
		return 0;

	BYTE *b = (BYTE*) SysInfo;
	b += SMB_INFO_CHASSIS_TYPE;

	wsprintf( outString, L"%u", *b );
	return 0;
}


DLLExports int WINAPI _smbchassislock( LPTSTR outString )
{
	if ( !outString )
		return -1;

	wcscpy_s( outString, MAX_ARG_LEN, StrNotAvailable );

	SMBStructure *SysInfo = FindStructure( SMB_TYPE_CHASSIS, -1 );
	if ( !SysInfo )
		return 0;

	if ( SysInfo->Length <= SMB_INFO_CHASSIS_TYPE )
		return 0;

	BYTE *b = (BYTE*) SysInfo;
	b += SMB_INFO_CHASSIS_TYPE;

	if ( *b & 0x80 )
		wcscpy_s( outString, MAX_ARG_LEN, L"Present" );
	else
		wcscpy_s( outString, MAX_ARG_LEN, L"Not present" );
	
	return 0;
}


DLLExports int WINAPI _smbchassistype( LPTSTR outString )
{
	if ( !outString )
		return -1;

	wcscpy_s( outString, MAX_ARG_LEN, StrNotAvailable );

	SMBStructure *SysInfo = FindStructure( SMB_TYPE_CHASSIS, -1 );
	if ( !SysInfo )
		return 0;

	if ( SysInfo->Length <= SMB_INFO_CHASSIS_TYPE )
		return 0;

	BYTE *b = (BYTE*) SysInfo;
	b += SMB_INFO_CHASSIS_TYPE;

	switch ( *b & 0x7f ) {
	case 0x01 :
		wcscpy_s( outString, MAX_ARG_LEN, L"Other" );
		break;

	case 0x03 :
		wcscpy_s( outString, MAX_ARG_LEN, L"Desktop" );
		break;

	case 0x04 :
		wcscpy_s( outString, MAX_ARG_LEN, L"Low Profile Desktop" );
		break;

	case 0x05 :
		wcscpy_s( outString, MAX_ARG_LEN, L"Pizza Box" );
		break;

	case 0x06 :
		wcscpy_s( outString, MAX_ARG_LEN, L"Mini Tower" );
		break;

	case 0x07 :
		wcscpy_s( outString, MAX_ARG_LEN, L"Tower" );
		break;

	case 0x08 :
		wcscpy_s( outString, MAX_ARG_LEN, L"Portable" );
		break;

	case 0x09 :
		wcscpy_s( outString, MAX_ARG_LEN, L"LapTop" );
		break;

	case 0x0a :
		wcscpy_s( outString, MAX_ARG_LEN, L"Notebook" );
		break;

	case 0x0b :
		wcscpy_s( outString, MAX_ARG_LEN, L"Hand Held" );
		break;

	case 0x0c :
		wcscpy_s( outString, MAX_ARG_LEN, L"Docking Station" );
		break;

	case 0x0d :
		wcscpy_s( outString, MAX_ARG_LEN, L"All in One" );
		break;

	case 0x0e :
		wcscpy_s( outString, MAX_ARG_LEN, L"Sub Notebook" );
		break;

	case 0x0f :
		wcscpy_s( outString, MAX_ARG_LEN, L"Space-saving" );
		break;

	case 0x10 :
		wcscpy_s( outString, MAX_ARG_LEN, L"Lunch Box" );
		break;

	case 0x11 :
		wcscpy_s( outString, MAX_ARG_LEN, L"Main Server Chassis" );
		break;

	case 0x12 :
		wcscpy_s( outString, MAX_ARG_LEN, L"Expansion Chassis" );
		break;

	case 0x13 :
		wcscpy_s( outString, MAX_ARG_LEN, L"SubChassis" );
		break;

	case 0x14 :
		wcscpy_s( outString, MAX_ARG_LEN, L"Bus Expansion Chassis" );
		break;

	case 0x15 :
		wcscpy_s( outString, MAX_ARG_LEN, L"Peripheral Chassis" );
		break;

	case 0x16 :
		wcscpy_s( outString, MAX_ARG_LEN, L"RAID Chassis" );
		break;

	case 0x17 :
		wcscpy_s( outString, MAX_ARG_LEN, L"Rack Mount Chassis" );
		break;

	case 0x18 :
		wcscpy_s( outString, MAX_ARG_LEN, L"Sealed-case PC" );
		break;

	case 0x19 :
		wcscpy_s( outString, MAX_ARG_LEN, L"Multi-system chassis" );
		break;

	case 0x1a :
		wcscpy_s( outString, MAX_ARG_LEN, L"Compact PCI" );
		break;

	case 0x1b :
		wcscpy_s( outString, MAX_ARG_LEN, L"Advanced TCA" );
		break;

	case 0x1c :
		wcscpy_s( outString, MAX_ARG_LEN, L"Blade" );
		break;

	case 0x1d :
		wcscpy_s( outString, MAX_ARG_LEN, L"Blade Enclosure" );
		break;

	default :
		wcscpy_s( outString, MAX_ARG_LEN, L"Unknown" );
	}

	return 0;
}


DLLExports int WINAPI _smbbiosdate( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *BiosInfo = FindStructure( SMB_TYPE_BIOSINFO, -1 );

	GetSMBString( BiosInfo, SMB_INFO_BIOS_DATE, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbbiosrev( LPTSTR outString )
{
	if ( !outString )
		return -1;

	wcscpy_s( outString, MAX_ARG_LEN, StrNotAvailable );

	SMBStructure *SysInfo = FindStructure( SMB_TYPE_BIOSINFO, -1 );
	if ( !SysInfo )
		return 0;

	if ( SysInfo->Length <= SMB_INFO_BIOS_REV_MINOR )
		return 0;

	BYTE *maj = (BYTE*) SysInfo;
	BYTE *min = (BYTE*) SysInfo;
	maj += SMB_INFO_BIOS_REV_MAJOR;
	min += SMB_INFO_BIOS_REV_MINOR;

	wsprintf( outString, L"%u.%u", *maj, *min );
	return 0;
}


DLLExports int WINAPI _smbbiosvendor( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *BiosInfo = FindStructure( SMB_TYPE_BIOSINFO, -1 );

	GetSMBString( BiosInfo, SMB_INFO_BIOS_VENDOR, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbbiosversion( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *BiosInfo = FindStructure( SMB_TYPE_BIOSINFO, -1 );

	GetSMBString( BiosInfo, SMB_INFO_BIOS_VERSION, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbcpuclock( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *CPUInfo = FindStructure( SMB_TYPE_CPU, -1 );

	wcscpy_s( outString, MAX_ARG_LEN, StrNotAvailable );

	if ( ( SMB_INFO_CPU_EXT_CLOCK + 1 ) >= CPUInfo->Length )
		return 0;


	BYTE *b = (BYTE*) CPUInfo;
	b += SMB_INFO_CPU_EXT_CLOCK;
	WORD *w = (WORD*) b;

	wsprintf( outString, L"%u", *w );
	return 0;
}


DLLExports int WINAPI _smbcpumfg( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *CPUInfo = FindStructure( SMB_TYPE_CPU, -1 );

	GetSMBString( CPUInfo, SMB_INFO_CPU_MANUFACTURER, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbcpusocket( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *CPUInfo = FindStructure( SMB_TYPE_CPU, -1 );

	GetSMBString( CPUInfo, SMB_INFO_CPU_SOCKET, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbcpuspeed( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *CPUInfo = FindStructure( SMB_TYPE_CPU, -1 );

	wcscpy_s( outString, MAX_ARG_LEN, StrNotAvailable );

	if ( ( SMB_INFO_CPU_SPEED + 1 ) >= CPUInfo->Length )
		return 0;


	BYTE *b = (BYTE*) CPUInfo;
	b += SMB_INFO_CPU_SPEED;
	WORD *w = (WORD*) b;

	wsprintf( outString, L"%u", *w );
	return 0;
}


DLLExports int WINAPI _smbcpuversion( LPTSTR outString )
{
	if ( !outString )
		return -1;

	SMBStructure *CPUInfo = FindStructure( SMB_TYPE_CPU, -1 );

	GetSMBString( CPUInfo, SMB_INFO_CPU_VERSION, outString, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI _smbver( LPTSTR outString )
{
	if ( !outString )
		return -1;


	if ( ( SMBVersionHigh >= 2 ) && ( SMBVersionHigh <= 10 ) )
		wsprintf( outString, L"%u.%u", SMBVersionHigh, SMBVersionLow );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StrNotAvailable );

	return 0;
}


DLLExports int WINAPI f_randx( LPTSTR String )
{
	if ( !String )
		return -1;


	unsigned int NumDigits = 8;
	int x = 0;

	if ( String[0] ) {
		wchar_t *ch = String;
		while ( iswspace( *ch ) )
			ch++;

		if ( ParseInt( ch, &NumDigits ) ) {
			_wcsupr_s( String, MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Invalid number: '%s'\r\n", ch );
			return -1;
		}

		if ( NumDigits > 1024 ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Number too large: %u\r\n", NumDigits );
			return -1;
		}
	}

	for ( unsigned int i = 0; i < NumDigits; i++ ) {
		x = rand() & 0x0f;
		if ( x < 10 )
			String[i] = '0' + x;
		else
			String[i] = 'A' + x - 10;
	}

	String[NumDigits] = 0x00;
	return 0;
}


DLLExports int WINAPI f_smbstring( LPTSTR String )
{
	if ( !String )
		return -1;


	unsigned int StructureType		= 0;
	unsigned int Offset				= 0;
	int Handle						= -1;
	SMBStructure *ThisStructure		= NULL;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	if ( arglen[0] ) {
		if ( ParseInt( arg[0], &StructureType ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad structure type \"%s\"\r\n", arg[0] );
			return -1;
		}

		if ( StructureType > 255 ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad structure type (must be 0 - 255) \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing structure type\r\n" );
		return -1;
	}


	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Offset) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad offset \"%s\"\r\n", arg[1] );
			return -1;
		}

		if ( ( Offset < 4 ) || ( Offset > 255 ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad offset (must be 4 - 255) \"%s\"\r\n", arg[1] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing offset\r\n" );
		return -1;
	}


	if ( arglen[2] ) {
		if ( ParseInt( arg[2], (unsigned int*) &Handle ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad handle \"%s\"\r\n", arg[2] );
			return -1;
		}

		if ( Handle > 0xffff ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad handle (must be 0x0000 - 0xffff) \"%s\"\r\n", arg[2] );
			return -1;
		}
	}
	else
		Handle = -1;


	wcscpy_s( String, MAX_ARG_LEN, StrNotAvailable );

	ThisStructure = FindStructure( StructureType, Handle );
	GetSMBString( ThisStructure, Offset, String, MAX_ARG_LEN );
	return 0;
}


DLLExports int WINAPI f_smbbyte( LPTSTR String )
{
	if ( !String )
		return -1;


	unsigned int StructureType		= 0;
	unsigned int Offset				= 0;
	int Handle						= -1;
	unsigned int Flags				= 0;
	SMBStructure *ThisStructure		= NULL;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	if ( arglen[0] ) {
		if ( ParseInt( arg[0], &StructureType ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad structure type \"%s\"\r\n", arg[0] );
			return -1;
		}

		if ( StructureType > 255 ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad structure type (must be 0 - 255) \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing structure type\r\n" );
		return -1;
	}


	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Offset) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad offset \"%s\"\r\n", arg[1] );
			return -1;
		}

		if ( ( Offset < 4 ) || ( Offset > 255 ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad offset (must be 4 - 255) \"%s\"\r\n", arg[1] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing offset\r\n" );
		return -1;
	}


	if ( arglen[2] ) {
		if ( ParseInt( arg[2], (unsigned int*) &Handle ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad handle \"%s\"\r\n", arg[2] );
			return -1;
		}

		if ( Handle > 0xffff ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad handle (must be 0x0000 - 0xffff) \"%s\"\r\n", arg[2] );
			return -1;
		}
	}
	else
		Handle = -1;


	if ( arglen[3] ) {
		if ( ParseInt( arg[3], &Flags ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad flags \"%s\"\r\n", arg[3] );
			return -1;
		}
	}


	wcscpy_s( String, MAX_ARG_LEN, StrNotAvailable );

	ThisStructure = FindStructure( StructureType, Handle );
	if ( !ThisStructure )
		return 0;

	if ( Offset >= ThisStructure->Length )
		return 0;


	BYTE *b = (BYTE*) ThisStructure;
	b += Offset;

	if ( Flags & 1 )
		wsprintf( String, L"0x%02x", *b );
	else
		wsprintf( String, L"%u", *b );

	return 0;
}


DLLExports int WINAPI f_smbword( LPTSTR String )
{
	if ( !String )
		return -1;


	unsigned int StructureType		= 0;
	unsigned int Offset				= 0;
	int Handle						= -1;
	unsigned int Flags				= 0;
	SMBStructure *ThisStructure		= NULL;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	if ( arglen[0] ) {
		if ( ParseInt( arg[0], &StructureType ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad structure type \"%s\"\r\n", arg[0] );
			return -1;
		}

		if ( StructureType > 255 ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad structure type (must be 0 - 255) \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing structure type\r\n" );
		return -1;
	}


	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Offset) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad offset \"%s\"\r\n", arg[1] );
			return -1;
		}

		if ( ( Offset < 4 ) || ( Offset > 254 ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad offset (must be 4 - 254) \"%s\"\r\n", arg[1] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing offset\r\n" );
		return -1;
	}


	if ( arglen[2] ) {
		if ( ParseInt( arg[2], (unsigned int*) &Handle ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad handle \"%s\"\r\n", arg[2] );
			return -1;
		}

		if ( Handle > 0xffff ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad handle (must be 0x0000 - 0xffff) \"%s\"\r\n", arg[2] );
			return -1;
		}
	}
	else
		Handle = -1;


	if ( arglen[3] ) {
		if ( ParseInt( arg[3], &Flags ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad flags \"%s\"\r\n", arg[3] );
			return -1;
		}
	}


	wcscpy_s( String, MAX_ARG_LEN, StrNotAvailable );

	ThisStructure = FindStructure( StructureType, Handle );
	if ( !ThisStructure )
		return 0;

	if ( ( Offset + 1 ) >= ThisStructure->Length )
		return 0;


	BYTE *b = (BYTE*) ThisStructure;
	b += Offset;
	WORD *w = (WORD*) b;

	if ( Flags & 1 )
		wsprintf( String, L"0x%04x", *w );
	else
		wsprintf( String, L"%u", *w );

	return 0;
}


DLLExports int WINAPI f_smbdword( LPTSTR String )
{
	if ( !String )
		return -1;


	unsigned int StructureType		= 0;
	unsigned int Offset				= 0;
	int Handle						= -1;
	unsigned int Flags				= 0;
	SMBStructure *ThisStructure		= NULL;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	if ( arglen[0] ) {
		if ( ParseInt( arg[0], &StructureType ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad structure type \"%s\"\r\n", arg[0] );
			return -1;
		}

		if ( StructureType > 255 ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad structure type (must be 0 - 255) \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing structure type\r\n" );
		return -1;
	}


	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Offset) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad offset \"%s\"\r\n", arg[1] );
			return -1;
		}

		if ( ( Offset < 4 ) || ( Offset > 252 ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad offset (must be 4 - 252) \"%s\"\r\n", arg[1] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing offset\r\n" );
		return -1;
	}


	if ( arglen[2] ) {
		if ( ParseInt( arg[2], (unsigned int*) &Handle ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad handle \"%s\"\r\n", arg[2] );
			return -1;
		}

		if ( Handle > 0xffff ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad handle (must be 0x0000 - 0xffff) \"%s\"\r\n", arg[2] );
			return -1;
		}
	}
	else
		Handle = -1;


	if ( arglen[3] ) {
		if ( ParseInt( arg[3], &Flags ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad flags \"%s\"\r\n", arg[3] );
			return -1;
		}
	}


	wcscpy_s( String, MAX_ARG_LEN, StrNotAvailable );

	ThisStructure = FindStructure( StructureType, Handle );
	if ( !ThisStructure )
		return 0;

	if ( ( Offset + 3 ) >= ThisStructure->Length )
		return 0;


	BYTE *b = (BYTE*) ThisStructure;
	b += Offset;
	DWORD *dw = (DWORD*) b;

	if ( Flags & 1 )
		wsprintf( String, L"0x%08x", *dw );
	else
		wsprintf( String, L"%u", *dw );

	return 0;
}


DLLExports int WINAPI f_smbqword( LPTSTR String )
{
	if ( !String )
		return -1;


	unsigned int StructureType		= 0;
	unsigned int Offset				= 0;
	int Handle						= -1;
	unsigned int Flags				= 0;
	SMBStructure *ThisStructure		= NULL;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	if ( arglen[0] ) {
		if ( ParseInt( arg[0], &StructureType ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad structure type \"%s\"\r\n", arg[0] );
			return -1;
		}

		if ( StructureType > 255 ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad structure type (must be 0 - 255) \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing structure type\r\n" );
		return -1;
	}


	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Offset) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad offset \"%s\"\r\n", arg[1] );
			return -1;
		}

		if ( ( Offset < 4 ) || ( Offset > 248 ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad offset (must be 4 - 248) \"%s\"\r\n", arg[1] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing offset\r\n" );
		return -1;
	}


	if ( arglen[2] ) {
		if ( ParseInt( arg[2], (unsigned int*) &Handle ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad handle \"%s\"\r\n", arg[2] );
			return -1;
		}

		if ( Handle > 0xffff ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad handle (must be 0x0000 - 0xffff) \"%s\"\r\n", arg[2] );
			return -1;
		}
	}
	else
		Handle = -1;


	if ( arglen[3] ) {
		if ( ParseInt( arg[3], &Flags ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad flags \"%s\"\r\n", arg[3] );
			return -1;
		}
	}


	wcscpy_s( String, MAX_ARG_LEN, StrNotAvailable );

	ThisStructure = FindStructure( StructureType, Handle );
	if ( !ThisStructure )
		return 0;

	if ( ( Offset + 7 ) >= ThisStructure->Length )
		return 0;


	BYTE *b = (BYTE*) ThisStructure;
	b += Offset;
	unsigned __int64 *qw = (unsigned __int64*) b;

	if ( Flags & 1 )
		wsprintf( String, L"0x%016I64x", *qw );
	else
		wsprintf( String, L"%I64u", *qw );

	return 0;
}


/*
DLLExports int WINAPI dumptable( LPTSTR inString )
{
	if ( !inString )
		return -1;


	if ( !FirmwareTableSize || !FirmwareTable ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"No firmware table!\r\n" );
		return -1;
	}


	int Width = 16;
	wchar_t String[SHORT_BUF_LEN] = L"";

	BYTE *Byte = FirmwareTable;

	for ( unsigned int i = 0; i < FirmwareTableSize; i++ ) {
		if ( ( i % Width ) == 0 )
			Printf( L"%04x: ", i );

		Printf( L"%02x ", Byte[i] );

		if ( Byte[i] >= 0x20 )
			String[i % Width] = Byte[i];
		else
			String[i % Width] = '.';

		if ( ( i % Width ) == ( Width - 1 ) ) {
			String[Width] = 0x00;
			Printf( L" %s\r\n", String );
		}
	}

	CrLf();

	return 0;
}
*/


DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t helpbuf[MAX_ARG_LEN] = L"";

	for ( int i = 0; i < NumHelpIndexItems; i++ ) {
		if ( !_wcsicmp( inString, HelpIndex[i].Name ) ) {
			wcscpy_s( helpbuf, MAX_ARG_LEN, HelpIndex[i].HelpText );

			if ( HelpIndex[i].UsageText )
				wcscat_s( helpbuf, MAX_ARG_LEN, HelpIndex[i].UsageText );

			MessageBox( GetForegroundWindow(), helpbuf, HelpWindowTitle, MB_OK );
			return 1;
		}
	}

	return 0;
}


/*
DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	for ( int i = 0; i < NumHelpIndexItems; i++ ) {
		if ( HelpIndex[i].UsageText && ( _wcsicmp( inString, HelpIndex[i].Name ) == 0 ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, HelpIndex[i].UsageText );
			return 1;
		}
	}

	return 0;
}
*/
